#ifndef MTRIANGULATE
#define MTRIANGULATE

#include "point.h"
#include <set>

enum PVertexType {Start, End, Left, Right, Merge, Split};

class PEdge;

class PVertex {
 public:
  PVertex (const Point &ip, PEdge *iedge = 0) : p(ip), edge(iedge) {}
  bool operator< (const PVertex &w) const
    { return p.y < w.p.y || p.y == w.p.y && p.x > w.p.x; }
  PVertexType type () const;
  void link (PVertex *w);
  void insert (PEdge *e);

  Point p;
  PEdge *edge;
};

typedef vector<PVertex *> PVertices;

class PEdgeOrder {
 public:
  bool operator() (PEdge *a, PEdge *b) const;
};

typedef set<PEdge *, PEdgeOrder> PEdgeSet;

typedef PEdgeSet::iterator PEdgeIter;

class PEdge {
 public:
  PEdge () : tail(0), helper(0), twin(0), next(0), flag(false), in(false) {}
  PEdge (PVertex *itail, PEdge *itwin = 0, PEdge *inext = 0, bool iflag = false)
    : tail(itail), helper(0), twin(itwin), next(inext), flag(iflag), in(false) {}
  bool left (const Point &p) const;
  void monotonic_loop (Points &pl, Points &pr);

  PVertex *tail, *helper;
  PEdge *twin, *next;
  bool flag, in;
  PEdgeIter node;
};

void triangulate (const Profile &region, Triangles &triangles);

void form_loops (const Profile &region, PVertices &vertices);

void form_loop (const Points &polygon, PVertices &vertices);

inline bool vertex_order (PVertex *v, PVertex *w) { return *w < *v; }

void make_y_monotonic (PVertices &vertices);

void handle_start (PVertex *v, PEdgeSet &sweep);

void handle_end (PVertex *v, PEdgeSet &sweep, PVertices &diagonals);

void handle_left (PVertex *v, PEdgeSet &sweep, PVertices &diagonals);

void handle_right (PVertex *v, PEdgeSet &sweep, PVertices &diagonals);

void handle_merge (PVertex *v, PEdgeSet &sweep, PVertices &diagonals);

void handle_split (PVertex *v, PEdgeSet &sweep, PVertices &diagonals);

PEdge * find_left (PVertex *v, PEdgeSet &sweep);

void check_diagonal (PVertex *v, PVertex *w, PVertices &diagonals);

void delete_loop (PVertices &vertices);

void triangulate_monotonic (PVertices &vertices, Triangles &triangles);

void triangulate_monotonic (const Points &pl, const Points &pr, 
			    Triangles &triangles);

void triangulate_add (const Point &p, Points &stack, bool rflag, 
		      Triangles &triangles);

void triangulate_pop (const Point &p, Points &stack, Triangles &triangles);

#endif
