#ifdef WIN32
#include <windows.h>
#define usleep Sleep
#endif

#include "system.h"
#include "postscript.h"
#include "plot.h"
#include <fstream>
#include "time.h"

System *Sys = 0;
double *Config = 0, AnimTime = 0.0, AnimSpeed = 1.0, 
  AnimStep = 1.0/30.0, CzoneWidth = 0.01;
int ZoomState = -1, PathState = -1;
Trajectory *Anim = 0;
bool Aflag = false, Fflag = true, Vflag = false, Mflag = false;
Point ZoomStart, ZoomEnd, PathStart, PathEnd;
Points Path;

SystemWindows Windows;

void graphics_init (int argc, char **argv, int wsize, System &s, Trajectory &t,
		    double *b, bool v)
{
  Sys = &s;
  Anim = &t;
  Vflag = v;
  s.form_profiles(b + 4);
  int n = s.parts.size();
  Config = new double [n];
  for (int i = 0; i < n; ++i)
    Config[i] = t[i+1];
  glutInit(&argc, argv);
  glutInitWindowSize(wsize, wsize);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  menu_init();
  int wp = glutCreateWindow("parts");
  Windows.push_back(SystemWindow(wsize, wsize, wp, b));
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyboard);	    
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutDisplayFunc(display);
  glutAttachMenu(GLUT_RIGHT_BUTTON);
  for (unsigned int i = 0; i < s.pairs.size(); ++i) {
    int wi = glutCreateWindow(cspace_window_name(s.pairs[i]).c_str());
    double bboxi[6];
    copy_bbox(s.pairs[i]->bbox, bboxi);
    Windows.push_back(SystemWindow(wsize, wsize, wi, bboxi, i));
    glutReshapeFunc(reshape);
    glutKeyboardFunc(keyboard);	    
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutDisplayFunc(display);
    glutAttachMenu(GLUT_RIGHT_BUTTON);
  }
  glutMainLoop();
}

void copy_bbox (double *b, double *c)
{
  for (int i = 0; i < 6; ++i)
    c[i] = b[i];
}

string cspace_window_name (Pair *pair)
{
  return pair->a->name + "/" + pair->b->name + " cspace";
}

SystemWindow & get_window ()
{
  int wid = glutGetWindow();
  for (unsigned int i = 0; i < Windows.size(); ++i)
    if (Windows[i].id == wid)
      return Windows[i];

  return Windows[0];

}

void reshape (int w, int h)
{
  SystemWindow &win = get_window();
  win.w = w;
  win.h = h;
  glViewport(0, 0, w, h);
}

void keyboard (unsigned char key, int x, int y)
{
  switch(key) {
  case 'a': 
    Aflag = !Aflag;
    break;
  case 'f':
    Fflag = !Fflag;
    break;
  case 'm':
    Mflag = !Mflag;
    break;
  case 'o':
    postscript_window();
    break;
  case 'n': 
    trajectory_step(AnimSpeed*AnimStep);
    break;
  case 'p':
    trajectory_step(- AnimSpeed*AnimStep);
    break;
  case 'r': {
    SystemWindow &w = get_window();
    copy_bbox(w.bboxz, w.bbox);
    break;
  }
  case 's':
    cerr << "Enter animation speed [current = " << AnimSpeed << "] ";
    cin >> AnimSpeed;
    break;
  case 'z':
    Vflag = !Vflag;
    if (Vflag && !Sys->pairs[0]->czone)
      Sys->form_czone(CzoneWidth);
    break;
  case 'w':
    cerr << "Enter contact zone width [current = " << CzoneWidth << "]";
    cin >> CzoneWidth;
    Sys->form_czone(CzoneWidth);
    break;
  case 'q': 
    exit(0);
  case 'P':
    PathState = PathState == -1 ? 0 : -1;
    break;
  }
  glutPostRedisplay();
}

void postscript_window ()
{
  cerr << "Enter filename: ";
  string fname;
  cin >> fname;
  ofstream ofstr(fname.c_str());
  SystemWindow &w = get_window();
  if (w.id == Windows[0].id) {
    PsView ps(144, 144, w.bbox, &ofstr);
    ps.draw_system(Sys, Config, 1.0, Fflag, 0.01);
  }
  else {
    Pair *pair = Sys->pairs[w.npair];
    //double asp = (w.bbox[3] - w.bbox[2])/(w.bbox[1] - w.bbox[0]);
    //PsView ps(108, (int) (asp*108), w.bbox, &ofstr);
    PsView ps(144, 144, w.bbox, &ofstr);
    ps.draw_cspace(Sys, pair, Config, Anim, w.bbox, Vflag);
  }
}

void mouse (int button, int buttonState, int x, int y)
{
  Point p = point(x, y);
  if (button == GLUT_LEFT_BUTTON)
    mouse_zoom(buttonState, p);
  else if (buttonState == GLUT_DOWN)
    return;
  else if (PathState > -1)
    mouse_path(p);
  else
    mouse_config(p);
}

Point point (int x, int y)
{
  SystemWindow &w = get_window();
  double u = x/(double) w.w, v = 1.0 - y/(double) w.h;
  double px = (1.0 - u)*w.bbox[0] + u*w.bbox[1];
  double py = (1.0 - v)*w.bbox[2] + v*w.bbox[3];
  return Point(px, py);
}

void mouse_zoom (int buttonState, const Point &p)
{
  if (buttonState == GLUT_DOWN) {
    ZoomStart = p;
    ZoomEnd = p;
    ZoomState = glutGetWindow();
  }
  else {
    SystemWindow &w = get_window();
    const Point &q = ZoomStart;
    w.bbox[0] = min(p.x, q.x);
    w.bbox[1] = max(p.x, q.x);
    w.bbox[2] = min(p.y, q.y);
    w.bbox[3] = max(p.y, q.y);
    ZoomState = -1;
  }
  glutPostRedisplay();
}

void mouse_config (const Point &p)
{
  SystemWindow &w = get_window();
  const Parts &pp = Sys->parts;
  const Pair &pair = *Sys->pairs[w.npair];
  int ia = find(pp.begin(), pp.end(), pair.a) - pp.begin(),
    ib = find(pp.begin(), pp.end(), pair.b) - pp.begin();
  Config[ia] = p.x;
  Config[ib] = p.y;
  glutPostRedisplay();
}

void mouse_path (const Point &p)
{
  switch (PathState) {
  case 0: case 2:
    PathStart = p;
    PathState = 1;
    break;
  case 1:
    PathEnd = p;
    Path.clear();
    int np = get_window().npair;
    if (np == -1)
      return;
    Pair *pair = Sys->pairs[np];
    pair->path(PathStart, PathEnd, Path, true);
    PathState = 2;
    break;
  }
  glutPostRedisplay();
}

void motion (int x, int y)
{
  if (ZoomState == -1) return;
  ZoomEnd = point(x, y);
  glutPostRedisplay();
}

void menu_init ()
{
  glutCreateMenu(menu);
  glutAddMenuEntry("Start animation", 1);
  glutAddMenuEntry("Next animation step", 2);
  glutAddMenuEntry("Previous animation step", 3);
  glutAddMenuEntry("Set animation speed", 4);
  glutAddMenuEntry("Draw parts in wireframe", 5);
  glutAddMenuEntry("Restore initial viewing box", 6);
  glutAddMenuEntry("Show contact zones", 7);
  glutAddMenuEntry("Set contact zone width", 8);
  glutAddMenuEntry("Save window in postscript", 9);
  glutAddMenuEntry("Exit", 10);
}

void menu (int value)
{
  switch (value) {
  case 1:
    Aflag = !Aflag;
    glutChangeToMenuEntry(1, Aflag ? "Stop animation" : "Start animation", 1);
    break;
  case 2:
    trajectory_step(AnimSpeed*AnimStep);
    break;
  case 3:
    trajectory_step(- AnimSpeed*AnimStep);
    break;
  case 4:
    cerr << "Enter animation speed [current = " << AnimSpeed << "] ";
    cin >> AnimSpeed;
    break;
  case 5:
    Fflag = !Fflag;
    glutChangeToMenuEntry(5, 
			  Fflag ? "Draw parts in wireframe" : "Draw parts filled",
			  5);
    break;
  case 6: {
    SystemWindow &w = get_window();
    copy_bbox(w.bboxz, w.bbox);
    break;
  }
  case 7:
    Vflag = !Vflag;
    glutChangeToMenuEntry(7, Vflag ? "Hide contact zones" : "Show contact zones", 
			  7);
    if (Vflag && !Sys->pairs[0]->czone)
      Sys->form_czone(CzoneWidth);
    break;
  case 8:
    cerr << "Enter contact zone width [current = " << CzoneWidth << "]";
    cin >> CzoneWidth;
    Sys->form_czone(CzoneWidth);
    break;
  case 9:
    postscript_window();
    break;
  case 10:
    exit(0);
  }
  glutPostRedisplay();
}

void display ()
{
  glEnableClientState(GL_VERTEX_ARRAY);
  clock_t tstart = clock();
  for (unsigned int i = 0; i < Windows.size(); ++i) {
    SystemWindow &w = Windows[i];
    glutSetWindow(w.id);
    if (i == 0)
      display_system(w.bbox);
    else
      display_cspace(w);
    if (ZoomState == w.id)
      display_zoom();
    glutSwapBuffers();
  }
  if (!Aflag) return;
  trajectory_step(AnimStep*AnimSpeed);
  if (AnimTime + AnimStep*AnimSpeed >= 
      Anim->at(Anim->size() - 1 - Sys->parts.size())) {
    Aflag = false;
    return;
  }
  glutPostRedisplay();
  double tsleep = 1e6*(AnimStep - (clock() - tstart));
  if (tsleep <= 0.0)
    return;
  usleep(int(tsleep));
}

void trajectory_step (double dt)
{
  static int i = 0;
  int n = 1 + Sys->parts.size(), m = Anim->size();
  double t = AnimTime + dt;
  if (t < 0.0) {
    t = Anim->at(m-n);
    i = m - 2*n;
  }
  else if (t >= Anim->at(m-n)) {
    t = 0.0;
    i = 0;
  }
  else if (dt > 0.0)
    while (i < m - n && t > Anim->at(i+n))
      i += n;
  else
    while (i > 0 && t < Anim->at(i))
      i -= n;
  AnimTime = t;
  double mu = (t - Anim->at(i))/(Anim->at(i+n) - Anim->at(i));
  for (int j = 0; j < n - 1; ++j)
    Config[j] = (1.0 - mu)*Anim->at(i+j+1) + mu*Anim->at(i+j+1+n);
}

void display_system (double *bbox)
{
  set_scale(bbox[0], bbox[1], bbox[2], bbox[3], 0.0, bbox[5] - bbox[4] + 2.0);
  glPushMatrix();
  glLoadIdentity();
  glTranslatef(0.0, 0.0, - bbox[5] - 1.0);
  if (Mflag && PathState == 2 && Sys->pairs.size() == 1)
    draw_path(Sys, Path);
  else
    draw_system(Sys, Config, Fflag);
  glPopMatrix();
}

void draw_path (System *system, const Points &path)
{
  double c[3] = {0, 0, 0};
  float white[3] = {1.0f, 1.0f, 1.0f};
  Pair *pair = system->pairs[0];
  bool rot = pair->a->rot;
  clear_screen(white);
  draw_part(pair->b, c, true);
  for (unsigned int i = 0; i < path.size(); ++i) {
    if (rot) {
      c[0] = - path[i].y;
      c[2] = path[i].x;
    }
    else {
      c[0] = path[i].x;
      c[1] = - path[i].y;
    }
    draw_part(pair->a, c, false);
  }
}

void display_cspace (SystemWindow &w)
{
  System *s = Sys;
  const Parts &pp = s->parts;
  const Pair &pair = *s->pairs[w.npair];
  draw_cspace(pair, w.bbox);
  int ia = find(pp.begin(), pp.end(), pair.a) - pp.begin(),
    ib = find(pp.begin(), pp.end(), pair.b) - pp.begin();
  Point p(Config[ia], Config[ib]);
  float red[3] = {1.0f, 0.0f, 0.0f};
  glLineWidth(2);
  draw_region(pair.gltraj, red, false);
  glColor3f(0.0f, 1.0f, 0.0f);
  glPointSize(5.0);
  draw_point(p);
  glColor3f(0.0f, 0.0f, 1.0f);
  switch (PathState) {
  case 1:
    draw_point(PathStart);
    break;
  case 2:
    draw_point(PathStart);
    draw_point(PathEnd);
    draw_polyline(Path);
  }
  glLineWidth(1);
}

void draw_polyline (const Points &points)
{
  glBegin(GL_LINE_STRIP);
  for (unsigned int i = 0; i < points.size(); ++i)
    glVertex2d(points[i].x, points[i].y);
  glEnd();
}

void draw_point (const Point &p)
{
  glBegin(GL_POINTS);
  glVertex2d(p.x, p.y);
  glEnd();
}

void display_zoom ()
{
  const Point &p = ZoomStart, &q = ZoomEnd;
  glColor3f(0.0f, 0.0f, 1.0f);
  glLineWidth(1);
  glBegin(GL_LINE_LOOP);
  glVertex2f(p.x, p.y);
  glVertex2f(p.x, q.y);
  glVertex2f(q.x, q.y);
  glVertex2f(q.x, p.y);
  glEnd();
}

void draw_system (System *s, double *c, bool fill)
{
  glEnable(GL_DEPTH_TEST);
  float white[3] = {1.0f, 1.0f, 1.0f};
  clear_screen(white);
  for (unsigned int i = 0; i < s->parts.size(); ++i) {
    double cp[3];
    Part *part = s->parts[i];
    part->config(c[i], cp);
    draw_part(part, cp, fill);
  }
  glDisable(GL_DEPTH_TEST);
}

void draw_part (Part *part, double *c, bool fill)
{
  glPushMatrix();
  glTranslatef(c[0], c[1], 0.0);
  glRotatef(c[2]*180.0/M_PI, 0.0, 0.0, 1.0);
  draw_regions(part->regions, part->color, fill);
  glPopMatrix();
}

void draw_cspace (const Pair &pair, double *bbox)
{
  float black[3] = {0.0f, 0.0f, 0.0f}, white[3] = {1.0f, 1.0f, 1.0f};
  float grey[3] = {.85f, .85f, .85f}, green[3] = {0.0f, 1.0f, 0.0f};
  set_scale(bbox[0], bbox[1], bbox[2], bbox[3], -1.0, 1.0);
  clear_screen(grey);
  draw_regions(pair.glregions, white, true);
  if (Vflag)
    draw_regions(pair.glczone, green, true);
  draw_axes(bbox, bbox + 2, black);
}

void set_scale (double x1, double x2, double y1, double y2,
		double z1, double z2)
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(x1, x2, y1, y2, z1, z2);
  glMatrixMode(GL_MODELVIEW);
}

void clear_screen (float *color)
{
  glClearColor(color[0], color[1], color[2], color[3]);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void draw_axes (const double *xr, const double *yr, float *color)
{
  glColor3fv(color);
  double x0, y0;
  x0 = xr[0] < 0 && xr[1] > 0 ? 0.0 : xr[0];
  y0 = yr[0] < 0 && yr[1] > 0 ? 0.0 : yr[0];
  glBegin(GL_LINES);
  glVertex2d(xr[0], y0);
  glVertex2d(xr[1], y0);
  glVertex2d(x0, yr[0]);
  glVertex2d(x0, yr[1]);
  glEnd();
}

void draw_regions (const glRegions &regions, float *color, bool fill)
{
  for (unsigned int i = 0; i < regions.size(); ++i)
    draw_region(*regions[i], color, fill);
}

void draw_region (const glRegion &r, float *color, bool fill)
{
  glColor3fv(color);
  if (fill) {
    glVertexPointer(3, GL_DOUBLE, 0, r.vt);
    glDrawArrays(GL_TRIANGLES, 0, r.nvt);
    glColor3f(0.0f, 0.0f, 0.0f);
  }
  glVertexPointer(3, GL_DOUBLE, 0, r.vb);
  for (unsigned int i = 0; i + 1 < r.nvb.size(); ++i)
    glDrawArrays(GL_LINE_STRIP, r.nvb[i], r.nvb[i+1] - r.nvb[i]);
}
