#include "point.h"

void Point::polar (double &r, double &a) const
{
  if (x == 0.0 && y == 0.0)
    r = a = 0.0;
  else {
    r = sqrt(x*x + y*y);
    a = atan2(y, x);
  }
}

bool pointInTriangle (const Point &p, const Point &a, const Point &b,
		      const Point &c)
{
  return signedArea(p, a, b) >= 0.0 && signedArea(p, b, c) >= 0.0 &&
    signedArea(p, c, a) >= 0.0;
}

double signedArea (const Points &polygon)
{
  double a = 0;
  int n = polygon.size();
  Point t = polygon[n-1];
  for (int i = 0; i < n; ++i) {
      a += t*polygon[i];
      t = polygon[i];
  }
  return 0.5*a;
}

Point random_point (double *bbox)
{
  return Point(random(bbox[0], bbox[1]), random(bbox[2], bbox[3]));
}

double random (double rmin, double rmax)
{
  return rmin + (rmax - rmin)*rand()/(double)RAND_MAX;
}

void read_points (Points &points, istream &istr)
{
  int n;
  double x, y;
  istr >> n;
  for (int i = 0; i < n; ++i) {
    istr >> x >> y;
    points.push_back(Point(x, y));
  }
}

void write_points (const Points &points, ostream &ostr)
{
  ostr << setprecision(16) << points.size() << endl;
  for (unsigned int i = 0; i < points.size(); ++i)
    ostr << points[i].x << " " << points[i].y << endl;
}

void write_profile (const Profile &profile, ostream &ostr)
{
  ostr << profile.size() << endl;
  for (unsigned int i = 0; i < profile.size(); ++i)
    write_points(profile[i], ostr);
}

bool Angle::in_arc (const Angle &s, const Angle &e) const
{
  bool cflag = s*e > 0.0;
  return cflag ? s*(*this) >= 0.0 && (*this)*e >= 0.0 :
    s*(*this) >= 0.0 || (*this)*e >= 0.0;
}

bool Angle::in_arc_strict (const Angle &s, const Angle &e) const
{
  return *this != s && *this != e && in_arc(s, e);
}

Angle Angle::midpoint (const Angle &e) const
{
  double x = 0.5*(c + e.c), y = 0.5*(s + e.s), d = sqrt(x*x + y*y);
  if (d == 0.0)
    return Angle(- s, c);
  if ((*this)*e < 0.0)
    d = - d;
  return Angle(x, y, d);
}

void write_triangles (const Triangles &triangles, ostream &ostr)
{
  for (unsigned int i = 0; i < triangles.size(); ++i) {
    const Triangle &tr = triangles[i];
    ostr << "((" << tr.a.x << " " << tr.a.y << ")("
	 << tr.b.x << " " << tr.b.y << ")("
	 << tr.c.x << " " << tr.c.y << "))";
  }
  cout << endl;
}
