#ifndef REGION
#define REGION
#include "point.h"

class RegIntersection {
 public:
  RegIntersection () {}
  RegIntersection (double it, const Point &ip, bool iin, int inp, int inb = 0)
    : t(it), p(ip), in(iin), np(inp), nb(inb) {}
  bool operator< (const RegIntersection &b) const 
    { return t < b.t || t == b.t && in && !b.in; }

  double t;
  Point p;
  bool in;
  int np, nb;
};

typedef vector<RegIntersection> RegIntersections;

class Region {
 public:
  Region () {}
  double area () const;
  Point interior_point () const;
  bool contains (const Point &p) const;
  double distance (const Point &p, int *edata, Point &q) const;
  void ray_intersections (const Point &p, const Point &q,
			  RegIntersections &res) const;
  bool ray_intersection (const Point &p, const Point &v,
			 double &t, Point &q, int *edata) const;
  bool path (const Point &p, const Point &q, Points &res) const;

  Profile boundary;
};

typedef vector<Region *> Regions;

void delete_regions (Regions *regions);

double point_boundary_distance (const Point &p, const Points &boundary, 
				int &ei, Point &q);

double point_segment_distance (const Point &p, const Point &a, const Point &b,
			       Point &q);

void ray_boundary_intersections (const Point &p, const Point &v, 
				 const Points &boundary, RegIntersections &res);

bool ray_line_intersection (const Point &p, const Point &v, const Point &a,
			    const Point &b, double &t, Point &q, bool &inflag);

#endif
