#include "system.h"

glRegion::glRegion (Profile &profile, double z1, double z2)
{
  int n = profile.size(), m = 0;
  for (int i = 0; i < n; ++i)
    m += profile[i].size();
  vb = new double [3*m];
  Triangles triangles;
  triangulate(profile, triangles);
  nvt = 3*triangles.size();
  vt = new double [3*nvt];
  double *vbp = vb, *vtp = vt;
  nvb.push_back(0);
  for (int i = 0; i < n; ++i) {
    nvb.push_back(nvb[i] + profile[i].size());
    for (unsigned int j = 0; j < profile[i].size(); ++j, vbp += 3) {
      vbp[0] = profile[i][j].x;
      vbp[1] = profile[i][j].y;
      vbp[2] = z2;
    }
  }
  double z = 0.5*(z1 + z2);
  for (unsigned int i = 0; i < triangles.size(); ++i, vtp += 9) {
      Triangle &t = triangles[i];
      vtp[0] = t.a.x;
      vtp[1] = t.a.y;
      vtp[2] = z;
      vtp[3] = t.b.x;
      vtp[4] = t.b.y;
      vtp[5] = z;
      vtp[6] = t.c.x;
      vtp[7] = t.c.y;
      vtp[8] = z;
  }
}

void delete_glRegions (glRegions &regions)
{
  for (unsigned int i = 0; i < regions.size(); ++i)
    delete regions[i];
  regions.clear();
}

Slice::~Slice ()
{
  delete_boundaries(boundaries);
}

void delete_boundaries (Boundaries &boundaries)
{
  for (unsigned int i = 0; i < boundaries.size(); ++i)
    delete_boundary(boundaries[i]);
}

Part::~Part ()
{
  for (unsigned int i = 0; i < slices.size(); ++i)
    delete slices[i];
  delete_glRegions (regions);
}

Regions * Part::cspace (Part *b, double acc, double *bbox, double *ibbox,
			bool usebbox) const
{
  if (usebbox)
    for (int i = 0; i < 4; ++i)
      bbox[i] = ibbox[i];
  else
    form_bbox(b, bbox, ibbox);
  RegionsList rlist;
  for (unsigned int i = 0; i < slices.size(); ++i)
    for (unsigned int j = 0; j < b->slices.size(); ++j)
      if (slices[i]->lb < b->slices[j]->ub &&
	  slices[i]->ub > b->slices[j]->lb) {
	Regions *reg = cspace_slice(slices[i]->boundaries, b->slices[j]->boundaries,
				    b, acc, bbox);
	if (reg)
	  rlist.push_back(reg);
      }
  if (rlist.empty())
    return 0;
  if (rlist.size() == 1)
    return rlist[0];
  Regions *res = intersection(rlist, bbox);
  for (unsigned int i = 0; i < rlist.size(); ++i)
    delete_regions(rlist[i]);
  return res;
}

void Part::form_bbox (Part *b, double *bbox, double *ibbox) const
{
  if (rot && b->rot) {
    bbox[0] = bbox[2] = - M_PI;
    bbox[1] = bbox[3] = M_PI;
  }
  else if (rot)
    form_bbox_rt(b, bbox);
  else
    form_bbox_tt(b, bbox);
  if (ibbox) {
    bbox[0] = max(bbox[0], ibbox[0]);
    bbox[1] = min(bbox[1], ibbox[1]);
    bbox[2] = max(bbox[2], ibbox[2]);
    bbox[3] = min(bbox[3], ibbox[3]);
  }
}

void Part::form_bbox_rt (Part *b, double *bbox) const
{
  bbox[0] = - M_PI;
  bbox[1] = M_PI;
  double d = rot_radius(), lbb, ubb;
  b->x_range(lbb, ubb);
  double w = 0.1*(2.0*d - lbb + ubb);
  bbox[2] = - d - ubb - w;
  bbox[3] = d - lbb + w;
}

void Part::form_bbox_tt (Part *b, double *bbox) const
{
  double albx, aubx, alby, auby, blbx, bubx, blby, buby;
  x_range(albx, aubx);
  y_range(alby, auby);
  b->x_range(blbx, bubx);
  b->y_range(blby, buby);
  double wx = 0.1*(bubx - albx - blbx + aubx);
  double wy = 0.1*(auby - blby - alby + buby);
  bbox[0] = blbx - aubx - wx;
  bbox[1] = bubx - albx + wx;
  bbox[2] = alby - buby - wy;
  bbox[3] = auby - blby + wy;
}

double Part::rot_radius () const
{
  double d = 0.0;
  for (unsigned int i = 0; i < slices.size(); ++i)
    for (unsigned int j = 0; j < slices[i]->boundaries.size(); ++j) {
      Boundary &b = *slices[i]->boundaries[j];
      for (unsigned int k = 0; k < b.size(); ++k) {
	d = max(d, b[k]->t.norm());
	if (b[k]->circle)
	  d = max(d, b[k]->o.norm()+b[k]->r);
      }
    }
  return d;
}
void Part::x_range (double &lb, double &ub) const
{
  lb = 1e10;
  ub = -1e10;
  for (unsigned int i = 0; i < slices.size(); ++i)
    for (unsigned int j = 0; j < slices[i]->boundaries.size(); ++j) {
      Boundary &b = *slices[i]->boundaries[j];
      for (unsigned int k = 0; k < b.size(); ++k) {
	double xt = b[k]->t.x;
	if (xt < lb) lb = xt;
	if (xt > ub) ub = xt;
	if (b[k]->circle) {
	  double xc = b[k]->o.x, r = b[k]->r;
	  if (xc - r < lb) lb = xc - r;
	  if (xc + r > ub) ub = xc + r;
	}
      }
    }
}

void Part::y_range (double &lb, double &ub) const
{
  lb = 1e10;
  ub = -1e10;
  for (unsigned int i = 0; i < slices.size(); ++i)
    for (unsigned int j = 0; j < slices[i]->boundaries.size(); ++j) {
      Boundary &b = *slices[i]->boundaries[j];
      for (unsigned int k = 0; k < b.size(); ++k) {
	double yt = b[k]->t.y;
	if (yt < lb) lb = yt;
	if (yt > ub) ub = yt;
	if (b[k]->circle) {
	  double yc = b[k]->o.y, r = b[k]->r;
	  if (yc - r < lb) lb = yc - r;
	  if (yc + r > ub) ub = yc + r;
	}
      }
    }
}

Regions * Part::cspace_slice (const Boundaries &ba, const Boundaries &bb,
			      Part *b, double acc, double *bbox) const
{
  Regions *regions = new Regions;
  Point oab = b->o - o;
  for (unsigned int i = 0; i < ba.size(); ++i)
    for (unsigned int j = 0; j < bb.size(); ++j)
      if (i == 0 || j == 0) {
	if (ba[i]->empty() || bb[j]->empty())
	  continue;
	Contacts contacts;
	boundary_contacts(ba[i], rot, bb[j], b->rot, oab, contacts);
	Segments curves;
	form_curves(rot, b->rot, contacts, acc, bbox, curves);
	Arrangement *arr = arrange(curves);
	form_regions(arr, b, ba[i], i == 0, bb[j], j == 0, acc, regions);
	for (unsigned int k = 0; k < contacts.size(); ++k)
	  delete contacts[k];
	delete arr;
	delete_segments(curves);
      }
  if (regions->empty()) {
    delete regions;
    return 0;
  }
  return regions;
}

void Part::form_regions (Arrangement *arr, Part *b, Boundary *ba, 
			 bool aout, Boundary *bb, bool bout, double acc,
			 Regions *regions) const
{
  Face *f = 0;
  bool flag = false;
  for (unsigned int i = 1; i < arr->faces.size(); ++i) {
    f = arr->faces[i];
    if (f->in) {
      Region *region = new Region;
      f->discretize(acc, region->boundary, EPSILON);
      flag = region->boundary[0].size() >= 3 && region->area() > MINAREA;
      if (flag) {
	Point p = region->interior_point();
	flag = !overlap(b, p.x, p.y, ba, aout, bb, bout);
      }
      if (flag) {
	regions->push_back(region);
	break;
      }
      else
	delete region;
    }
  }
  if (!flag) 
    return;
  arr->set_crossing_numbers(f, 0);
  for (unsigned int i = 1; i < arr->faces.size(); ++i) {
    Face *g = arr->faces[i];
     if (g != f && g->cn == 0 && g->in) {
       Region *region = new Region;
       g->discretize(acc, region->boundary, EPSILON);
       regions->push_back(region);
    }
  }
}

bool Part::overlap (Part *b, double u, double v, Boundary *ba, bool aout,
		    Boundary *bb, bool bout) const
{
  Angle th(rot ? u : 0.0), thb(b->rot ? v : 0.0);
  Point d = rot ? Point(0.0, 0.0) : Point(u, 0.0);
  Point db = b->rot ? b->o - o : (rot ? Point(v, 0.0) : Point(0.0, v));
  Boundary *mba = ::move(ba, th, d);
  Boundary *mbb = ::move(bb, thb, db);
  bool res = ::overlap(mba, mbb);
  if (!res) 
    if (aout && bout) {
      Point &pa = mba->at(0)->t;
      res = contains(*mbb, pa);
      if (!res) {
	Point &pb = mbb->at(0)->t;
	res = contains(*mba, pb);
      }
    }
    else if (aout && !bout) {
      Point &pa = mba->at(0)->t;
      res = !contains(*mbb, pa);
    }
    else {
      Point &pb = mbb->at(0)->t;
      res = !contains(*mba, pb);
    }
  delete_boundary(mba);
  delete_boundary(mbb);
  return res;
}

Boundary * move (Boundary *b, const Angle &a, const Point &d)
{
  Boundary *mb = new Boundary;
  for (unsigned int i = 0; i < b->size(); ++i) {
    Feature *mf = new Feature(b->at(i)->move(a, d));
    mb->push_back(mf);
  }
  return mb;
}

bool overlap (Boundary *b1, Boundary *b2)
{
  for (unsigned int i = 0; i < b1->size(); ++i)
    for (unsigned int j = 0; j < b2->size(); ++j)
      if (b1->at(i)->intersects(*b2->at(j)))
	return true;
  return false;
}

bool contains (const Boundary &boundary, Point p)
{
  for (unsigned int i = 0; i < boundary.size(); ++i)
    if (boundary[i]->t.x == p.x || boundary[i]->h.x == p.x)
      p.x += EPSILON;
  bool in = false;
  for (unsigned int i = 0; i < boundary.size(); ++i)
    if (boundary[i]->above(p))
      in = !in;
  return in;
}

/*
bool Part::overlap (Part *b, double u, double v) const
{
  Angle th(rot ? u : 0.0), thb(b->rot ? v : 0.0);
  Point d = rot ? Point(0.0, 0.0) : Point(u, 0.0);
  Point db = b->rot ? b->o - o : (rot ? Point(v, 0.0) : Point(0.0, v));
  for (unsigned int i = 0; i < slices.size(); ++i) {
    Slice *s = slices[i];
    for (unsigned int ib = 0; ib < b->slices.size(); ++ib) {
      Slice *sb = b->slices[ib];
      if (s->lb < sb->ub && s->ub > sb->lb &&
	  ::overlap(s->boundaries, th, d, sb->boundaries, thb, db))
	return true;
    }
  }
  return false;
}

bool overlap (const Boundaries &boundaries1, const Angle &a1, const Point &d1,
	      const Boundaries &boundaries2, const Angle &a2, const Point &d2)
{
  Boundaries rb1, rb2;
  move(boundaries1, a1, d1, rb1);
  move(boundaries2, a2, d2, rb2);
  bool res = overlap1(rb1, rb2) || overlap2(rb1, rb2) || overlap2(rb2, rb1);
  delete_boundaries(rb1);
  delete_boundaries(rb2);
  return res;
}

void move (const Boundaries &boundaries, const Angle &a, const Point &d,
	   Boundaries &res)
{
  for (unsigned int i = 0; i < boundaries.size(); ++i) {
    Boundary &b = *boundaries[i], *rb = new Boundary;
    res.push_back(rb);
    for (unsigned int j = 0; j < b.size(); ++j) {
      Feature *f = b[j], *rf = new Feature;
      rb->push_back(rf);
      *rf = f->move(a, d);
    }
  }
}

bool overlap1 (const Boundaries &boundaries1, const Boundaries &boundaries2)
{
  for (unsigned int i1 = 0; i1 < boundaries1.size(); ++i1)
    for (unsigned int i2 = 0; i2 < boundaries2.size(); ++i2) {
      Boundary &b1 = *boundaries1[i1], &b2 = *boundaries2[i2];
      for (unsigned int j1 = 0; j1 < b1.size(); ++j1)
	for (unsigned int j2 = 0; j2 < b2.size(); ++j2)
	  if (b1[j1]->intersects(*b2[j2]))
	    return true;
    }
  return false;
}

bool overlap2 (const Boundaries &boundaries1, const Boundaries &boundaries2)
{
  Point &p2 = (*boundaries2[0])[0]->t;
  if (!contains(*boundaries1[0], p2))
    return false;
  for (unsigned int i = 1; i < boundaries1.size(); ++i)
    if (contains(*boundaries1[i], p2))
      return false;
  return true;
}
*/

void Part::read (istream &istr)
{
  int nslices, dc[4];
  double pars[7];
  istr >> name >> nslices >> dc[0] >> dc[1] >> dc[2] >> dc[3];
  for (int i = 0; i < 3; ++i)
    color[i] = dc[i]/255.0f;
  for (int i = 0; i < 7; ++i)
    istr >> pars[i];
  rot = pars[0] == 3.0;
  o.x = pars[1];
  o.y = pars[2];
  if (!rot) {
    if (pars[0] == 1.0 || pars[0] == 5.0)  
      t = Angle(1.0, 0.0);
    else if (pars[0] == 2.0)
      t = Angle(0.0, 1.0);
    else if (pars[0] == 4.0)
      t = Angle(pars[6]);
    o.x -= (t.c*o.x + t.s*o.y)*t.c;
    o.y -= (t.c*o.x + t.s*o.y)*t.s;
  }
  Angle th(rot ? 0.0 : pars[3]);
  for (int i = 0; i < nslices; ++i) {
    Slice *slice = read_slice(istr, th);
    slices.push_back(slice);
  }  
}

Slice * Part::read_slice (istream &istr, const Angle &th)
{
  int nboundaries;
  double lb, ub;
  istr >> nboundaries >> lb >> ub;
  Slice *slice = new Slice(lb, ub);
  for (int i = 0; i < nboundaries; ++i)
    slice->boundaries.push_back(read_boundary(istr, th, i == 0));
  return slice;
}

Boundary * Part::read_boundary (istream &istr, const Angle &th, bool flag) const
{
  Boundary segments;
  int ngroups;
  istr >> ngroups;
  for (int i = 0; i < ngroups; ++i) {
    Boundary sgroup;
    read_sgroup(istr, th, sgroup);
    for (unsigned int j = 0; j < sgroup.size(); ++j)
      segments.push_back(sgroup[j]);
  }
  if (!flag)
    reverse(segments.begin(), segments.end());
  Boundary *boundary = fillet(segments);
  return boundary;
}
 
void Part::read_sgroup (istream &istr, const Angle &th, Boundary &sgroup) const
{
  Boundary segments;
  int nsegments, noffsets;
  istr >> nsegments;
  for (int i = 0; i < nsegments; ++i)
    segments.push_back(read_feature(istr, th));
  istr >> noffsets;
  double *offsets = new double [noffsets];
  for (int i = 0; i < noffsets; ++i)
    istr >> offsets[i];
  for (int i = 0; i < noffsets; ++i)
    for (int j = 0; j < nsegments; ++j)
      sgroup.push_back(shift_feature(segments[j], offsets[i]));
  delete [] offsets;
}

Feature * Part::shift_feature (Feature *f, double u) const
{
  if (u == 0.0)
    return f;
  Feature *sf = new Feature;
  if (rot)
    *sf = f->move(u, 0.0, 0.0);
  else
    *sf = f->move(0.0, u*t.c, u*t.s);
  return sf;
}  

Part * Part::move (const Angle &a, const Point &d) const
{
  Part *res = new Part;
  res->name = name;
  for (int i = 0; i < 3; ++i)
    res->color[i] = color[i];
  res->rot = rot;
  res->o = o;
  res->t = t;
  for (unsigned int i = 0; i < slices.size(); ++i) {
    Slice *s = new Slice(slices[i]->lb, slices[i]->ub);
    res->slices.push_back(s);
    for (unsigned int j = 0; j < slices[i]->boundaries.size(); ++j) {
      Boundary *b = new Boundary;
      s->boundaries.push_back(b);
      for (unsigned int k = 0; k < slices[i]->boundaries[j]->size(); ++k) {
	Feature *f = new Feature;
	*f = (*slices[i]->boundaries[j])[k]->move(a, d);
	b->push_back(f);
      }
    }
  }
  return res;
}

Part * Part::offset (double d) const
{
  Part *p = new Part;
  p->name = name;
  for (int i = 0; i < 3; ++i)
    p->color[i] = color[i];
  p->rot = rot;
  p->o = o;
  p->t = t;
  for (unsigned int i = 0; i < slices.size(); ++i) {
    Slice *slice = new Slice(slices[i]->lb, slices[i]->ub);
    p->slices.push_back(slice);
    for (unsigned int j = 0; j < slices[i]->boundaries.size(); ++j)
      slice->boundaries.push_back(::offset(slices[i]->boundaries[j], d));
  }
  return p;
}

bool Part::wrap ()
{
  if (!rot) 
    return false;
  if (v > 0.0 && fabs(x - M_PI) < EPSILON) {
    x = - M_PI;
    return true;
  }
  if (v < 0.0 && fabs(x + M_PI) < EPSILON) {
    x = M_PI;
    return true;
  }
  return false;
}

void Part::config (double mp, double *c) const
{
  if (rot) {
    c[0] = o.x;
    c[1] = o.y;
    c[2] = mp;
  }
  else {
    c[0] = t.c*mp + o.x;
    c[1] = t.s*mp + o.y;
    c[2] = 0.0;
  }
}

void Part::form_profiles (double acc, double *z)
{
  for (unsigned int i = 0; i < slices.size(); ++i) {
    z[0] = min(z[0], slices[i]->lb);
    z[1] = max(z[1], slices[i]->ub);
    profiles.push_back(Profile());
    Profile &profile = profiles[profiles.size()-1];
    for (unsigned int j = 0; j < slices[i]->boundaries.size(); ++j) {
      profile.push_back(Points());
      Points &points = profile[profile.size()-1];
      Boundary &boundary = *slices[i]->boundaries[j];
      for (unsigned int k = 0; k < slices[i]->boundaries[j]->size(); ++k)
	boundary[k]->discretize(acc, points);
    }
  }
}

void Part::triangulate ()
{
  for (unsigned int i = 0; i < profiles.size(); ++i)
    regions.push_back(new glRegion(profiles[i], slices[i]->lb, slices[i]->ub));
}
			
void Part::triangulate_3d (ostream &ostr) const
{
  vector<double> pts;
  for (unsigned int i = 0; i < slices.size(); ++i)
    triangulate_profile_3d(profiles[i], slices[i]->lb, slices[i]->ub, pts);
  write_triangles_vtk(ostr, pts);
}

void triangulate_profile_3d (const Profile &profile, double lb, double ub,
			     vector<double> &pts)
{
  Triangles triangles;
  triangulate(profile, triangles);
  triangulate_profile_tb(triangles, lb, false, pts);
  triangulate_profile_tb(triangles, ub, true, pts);
  triangulate_profile_s(profile, lb, ub, pts);
}

void triangulate_profile_tb (const Triangles &triangles, double z, bool ccw, 
			     vector<double> &pts)
{
  for (unsigned int i = 0; i < triangles.size(); ++i) {
    const Triangle &tr = triangles[i];
    Point ptr[3] = {tr.a, tr.b, tr.c};
    double sa = signedArea(tr.a, tr.b, tr.c);
    bool flip = ccw ? sa < 0.0 : sa > 0.0;
    for (int j = 0; j < 3; ++j) {
      Point &p = flip ? ptr[2-j] : ptr[j];
      add_point_3d(p.x, p.y, z, pts);
    }
  }
}

void triangulate_profile_s (const Profile &profile, double lb, double ub,
			    vector<double> &pts)
{
  for (unsigned int i = 0; i < profile.size(); ++i) {
    const Points &points = profile[i];
    int n = points.size();
    Point a = points[n-1];
    for (int j = 0; j < n; ++j) {
      Point b = points[j];
      if (a != b) {
	add_point_3d(a.x, a.y, lb, pts);
	add_point_3d(b.x, b.y, lb, pts);
	add_point_3d(b.x, b.y, ub, pts);
	add_point_3d(a.x, a.y, lb, pts);
	add_point_3d(b.x, b.y, ub, pts);
	add_point_3d(a.x, a.y, ub, pts);
      }
      a = b;
    }
  }
}

void add_point_3d (double x, double y, double z, vector<double> &pts)
{
  pts.push_back(x);
  pts.push_back(y);
  pts.push_back(z);
}

void write_triangles_vtk (ostream &ostr, vector<double> &pts)
{
  int nt = pts.size()/9, np = 3*nt, m = 4*nt;
  ostr << "# vtk DataFile Version 3.0" << endl
       << "vtk output" << endl << "ASCII" << endl
       << "DATASET POLYDATA" << endl 
       << "POINTS " << np << " float" << endl;
  for (int i = 0; i < nt; ++i) {
    for (int j = 0; j < 9; ++j)
      ostr << pts[9*i+j] << " ";
    ostr << endl;
  }
  ostr << endl << "POLYGONS " << nt << " " << m << endl;
  for (int i = 0; i < np; i += 3)
    ostr << "3 " << i << " " << i + 1 << " " << i + 2 << endl;
  ostr << endl << "CELL_DATA " << nt << endl 
       << "POINT_DATA " << np << endl;
}

void triangulate_fa (Part *a, Part *b, double acc, double *bbox,
		     int axis, double *aint, vector<double> &pts)
{
  double fbox[4];
  Regions *free = a->cspace(b, acc, fbox, bbox, true);
  Regions *blocked = complement(free, bbox);
  vector<double> apts;
  for (unsigned int i = 0; i < blocked->size(); ++i) {
    Triangles triangles;
    triangulate_regions(blocked, triangles);
    extrude(triangles, axis, aint, apts);
  }
  transpose(apts, axis, pts);
  delete_regions(free);
  delete_regions(blocked);
}

void extrude (const Triangles &triangles, int axis, double *aint,
	      vector<double> &pts)
{
  for (unsigned int i = 0; i < triangles.size(); ++i) {
    const Triangle &tr = triangles[i];
    Point trp[4] = {tr.a, tr.b, tr.c, tr.a};
    for (int j = 0; j < 3; ++j) {
      add_point_3d(trp[j].x, trp[j].y, aint[0], pts);
      add_point_3d(trp[j+1].x, trp[j+1].y, aint[0], pts);
      add_point_3d(trp[j+1].x, trp[j+1].y, aint[1], pts);
      add_point_3d(trp[j].x, trp[j].y, aint[0], pts);
      add_point_3d(trp[j+1].x, trp[j+1].y, aint[1], pts);
      add_point_3d(trp[j].x, trp[j].y, aint[1], pts);
    }
    for (int k = 0; k < 2; ++k)
      for (int j = 0; j < 3; ++j)
	add_point_3d(trp[j].x, trp[j].y, aint[k], pts);
  }
}

void transpose (const vector<double> &apts, int axis, vector<double> &pts)
{
  int p[3] = {0, 1, 2};
  if (axis == 0) {
    p[0] = 2;
    p[1] = 0;
    p[2] = 1;
  }
  else if (axis == 1) {
    p[1] = 2;
    p[2] = 1;
  }
  for (unsigned int i = 0; i < apts.size()/3; ++i)
    for (int j = 0; j < 3; ++j)
      pts.push_back(apts[3*i+p[j]]);
}

void triangulate_patches (const Boundary &a, const Boundary &b,
			  double dt, double acc, int k, ostream &ostr)
{
  vector<double> pts;
  for (unsigned int i = 0; i < a.size(); ++i)
    for (unsigned int j = 0; j < b.size(); ++j)
      triangulate_patch(a[i], b[j], dt, acc, k, pts);
  write_triangles_vtk(ostr, pts);
}

void triangulate_patch (Feature *f, Feature *g, double dt, double acc,
			int k, ostream &ostr)
{
  vector<double> pts;
  triangulate_patch(f, g, dt, acc, k, pts);
  write_triangles_vtk(ostr, pts);
}

void triangulate_patch (Feature *f, Feature *g, double dt, double acc,
			int k, vector<double> &pts)
{
  if (f->circle && g->circle)
    triangulate_patch_aa(f, g, dt, acc, k, pts);
  else if (f->circle)
    triangulate_patch_al(f, g, dt, k, pts);
  else
    triangulate_patch_la(f, g, dt, k, pts);
}

void triangulate_patch_aa (Feature *arc1, Feature *arc2, double dt, double acc,
			   int k, vector<double> &pts)
{
  if (arc1->r == 0.0 && arc2->r == 0.0 ||
      arc1->r == 0.0 && !arc1->in ||
      arc2->r == 0.0 && !arc2->in)
    return;
  if (arc1->r > 0.0 && arc2->r > 0.0) {
    double c1 = (arc1->in ? 1.0 : -1.0)/arc1->r;
    double c2 = (arc2->in ? 1.0 : -1.0)/arc2->r;
    if (c1 + c2 <= 0.0)
      return;
  }
  Feature parts1[2], parts2[2];
  int n1 = arc1->split(parts1), n2 = arc2->split(parts2);
  for (int i = 0; i < n1; ++i)
    for (int j = 0; j < n2; ++j)
      triangulate_patch_aa_aux(parts1 + i, parts2 + j, dt, acc, k, pts);
}    

void triangulate_patch_aa_aux (Feature *arc1, Feature *arc2, double dt, double acc,
			       int k, vector<double> &pts)
{
  Angle s1 = arc1->in ? arc1->nt : arc1->nh;
  Angle e1 = arc1->in ? arc1->nh : arc1->nt;
  Angle s2 = arc2->in ? arc2->nt : arc2->nh;
  Angle e2 = arc2->in ? arc2->nh : arc2->nt;
  double th = mod_pi(s2.angle() - e1.angle() + M_PI, true);
  double the = mod_pi(e2.angle() - s1.angle() + M_PI);
  if (the <= th)
    the += 2.0*M_PI;
  double t = arc1->aa_r(*arc2);
  Points points;
  points.push_back(arc2->o - Angle(th)*arc1->o + t*s2);
  while (th < the) {
    double nth = min(th + dt, the);
    if (th < M_PI && nth > M_PI)
      nth = M_PI;
    Points npoints;
    if (nth == the)
      npoints.push_back(arc2->o - Angle(nth)*arc1->o + t*e2);
    else {
      Angle ns1 = s1 + (nth - M_PI), ne1 = e1 + (nth - M_PI);
      Angle s = ns1.in_arc(s2, e2) ? ns1 : s2;
      Angle e = ne1.in_arc(s2, e2) ? ne1 : e2;
      Point o = arc2->o - Angle(nth)*arc1->o;
      Feature nsum(o+t*s, o+t*e, o, t, arc1->in && arc2->in);
      nsum.discretize(acc, npoints);
    }
    triangulate_strip(points, mod_pi(th, true), npoints, mod_pi(nth), k, pts);
    points = npoints;
    th = nth;
  }
}

void triangulate_strip (Points &lb, double lth, Points &ub, double uth,
			int k, vector<double> &pts)
{
  Points slb, sub;
  int n = k*max(lb.size(), ub.size());
  resample(lb, n, slb);
  resample(ub, n, sub);
  for (int i = 0; i < n - 1; ++i) {
    add_point_3d(slb[i].x, slb[i].y, lth, pts);
    add_point_3d(slb[i+1].x, slb[i+1].y, lth, pts);
    add_point_3d(sub[i].x, sub[i].y, uth, pts);
    add_point_3d(slb[i+1].x, slb[i+1].y, lth, pts);
    add_point_3d(sub[i+1].x, sub[i+1].y, uth, pts);
    add_point_3d(sub[i].x, sub[i].y, uth, pts);
  }
}

void resample (const Points &points, int n, Points &res)
{
  double d = 0.0;
  int m = points.size() - 1;
  for (int i = 0; i < m; ++i)
    d += points[i].distance(points[i+1]);
  d /= n - 1.0;
  int i = 0;
  Point p = points[0];
  res.push_back(p);
  while (i < m && (int) res.size() + 1 < n) {
     double e = d;
    while (true) {
      double f = p.distance(points[i+1]);
      if (e < f) {
	p = p + (e/f)*(points[i+1] - p);
	break;
      }
      ++i;
      e -= f;
      p = points[i];
    }
    res.push_back(p);
  }
  res.push_back(points[m]);
}

void triangulate_patch_al (Feature *arc, Feature *line, double dt,
			   int k, vector<double> &pts)
{
  if (!arc->in) return;
  double th, the;
  if (arc->full_circle()) {
    th = - M_PI;
    the = M_PI;
  }
  else {
    th = mod_pi(arc->nt.angle() - line->nt.inv().angle(), true);
    the = mod_pi(arc->nh.angle() - line->nt.inv().angle());
    if (the < th)
      the += 2.0*M_PI;
  }
  Point e = arc->r*line->nt + line->t;
  Point f = arc->r*line->nt + line->h;
  Point a = e - Angle(th)*arc->o, b = f - Angle(th)*arc->o;
  while (th < the) {
    double nth = min(th + dt, the);
    if (th < M_PI && nth > M_PI)
      nth = M_PI;
    Angle anth(nth);
    Point na = e - anth*arc->o, nb = f - anth*arc->o;
    double thm = mod_pi(th, true);
    Points lb, ub;
    lb.push_back(a);
    lb.push_back(b);
    ub.push_back(na);
    ub.push_back(nb);
    triangulate_strip(lb, thm, ub, nth, k, pts);
    th = nth;
    a = na;
    b = nb;
  }
}

void triangulate_patch_la (Feature *line, Feature *arc, double dt,
			   int k, vector<double> &pts)
{
  if (!arc->in) return;
  double th, the;
  if (arc->full_circle()) {
    th = - M_PI;
    the = M_PI;
  }
  else {
    th = mod_pi(arc->nt.angle() - line->nt.inv().angle(), true);
    the = mod_pi(arc->nh.angle() - line->nt.inv().angle());
    if (the < th)
      the += 2.0*M_PI;
  }
  Point e = arc->r*line->nt + line->t;
  Point f = arc->r*line->nt + line->h;
  Point a = arc->o - Angle(th)*e, b = arc->o - Angle(th)*f;
  while (th < the) {
    double nth = min(th + dt, the);
    if (th < M_PI && nth > M_PI)
      nth = M_PI;
    Angle anth(nth);
    Point na = arc->o - anth*e, nb = arc->o - anth*f;
    double thm = mod_pi(th, true);
    Points lb, ub;
    lb.push_back(a);
    lb.push_back(b);
    ub.push_back(na);
    ub.push_back(nb);
    triangulate_strip(lb, thm, ub, nth, k, pts);
    th = nth;
    a = na;
    b = nb;
  }
}

void sampled_cspace (Part *a, Part *b, double ths, double the, 
		     double dt, double acc, int k, ostream &ostr)
{
  vector<double> pts;
  Point o;
  double bbox[4];
  double th = ths;
  Part *ra = a->move(th, o);
  Regions *reg = ra->cspace(b, acc, bbox, 0);
  while (th < the) {
    double nth = min(th + dt, M_PI);
    delete ra;
    ra = a->move(nth, o);
    Regions *nreg = ra->cspace(b, acc, bbox, 0);
    int m = min(reg->size(), nreg->size());
    for (int i = 0; i < m; ++i) {
      int j = i == 0 ? 1 : 0;
      Points rb0, nrb0;
      reflect_points(reg->at(i)->boundary[j], rb0);
      reflect_points(nreg->at(i)->boundary[j], nrb0);
      triangulate_strip(rb0, th, nrb0, nth, k, pts);
    }
    delete_regions(reg);
    th = nth;
    reg = nreg;
  }
  delete ra;
  delete_regions(reg);
  write_triangles_vtk(ostr, pts);
}

void sliced_cspace (Part *a, Part *b, double ths, double the, 
		    double dt, double acc, ostream &ostr)
{
  Point o;
  double bbox[4];
  double th = ths;
  while (th < the) {
    Part *ra = a->move(th, o);
    Regions *reg = ra->cspace(b, acc, bbox, 0);
    ostr << reg->size() << endl;
    for (unsigned int i = 0; i < reg->size(); ++i) {
      Profile prof;
      reflect_profile(reg->at(i)->boundary, prof);
      write_profile(prof, ostr);
    }
    delete ra;
    delete_regions(reg);
    th = min(th + dt, M_PI);
  }
}

void reflect_points (const Points &pts, Points &rpts)
{
  for (unsigned int i = 0; i < pts.size(); ++i) {
    const Point &p = pts[i];
    rpts.push_back(Point(p.x, - p.y));
  }
}

void reflect_profile (const Profile &prof, Profile &rprof)
{
  for (unsigned int i = 0; i < prof.size(); ++i) {
    Points pts;
    reflect_points(prof[i], pts);
    rprof.push_back(pts);
  }
}

Feature * read_feature (istream &istr, const Angle &th)
{
  int index, circle;
  Point t, h;
  bool in;
  istr >> index >> circle >> t >> h >> in;
  if (!circle)
    return new Feature(th*t, th*h, in);
  Point o;
  double r, sa, ea;
  istr >> o >> r >> sa >> ea;
  return new Feature(th*t, th*h, th*o, r, in);
}

Regions * intersection (const RegionsList &rlist, double *bbox)
{
  int n = rlist.size();
  for (int i = 0; i < n; ++i)
    if (!rlist[i])
      return 0;
  Segments segments;
  for (int i = 0; i < n; ++i) {
    Regions *ri = rlist[i];
    for (unsigned int j = 0; j < ri->size(); ++j)
      region_segments(ri->at(j), bbox, segments);
  }
  bool brot = bbox[2] == - M_PI && bbox[3] == M_PI;
  form_curves_bbox(brot, bbox, segments);
  Arrangement *arr = arrange(segments);
  Face *f0 = 0;
  Point p0;
  for (unsigned int i = 1; i < arr->faces.size(); ++i) {
    Face *fi = arr->faces[i];
    Region ri;
    fi->traverse(ri.boundary);
    if (ri.boundary[0].size() >= 3 && ri.area() > MINAREA) {
      f0 = fi;
      p0 = ri.interior_point();
      break;
    }
  }
  if (f0 == 0)
    return 0;
  int cn0 = 0;
  for (int i = 0; i < n; ++i)
    for (unsigned int j = 0; j < rlist[i]->size(); ++j)
      if (rlist[i]->at(j)->contains(p0)) {
	++cn0;
	break;
      }
  arr->set_crossing_numbers(f0, cn0);
  Regions *regions = new Regions;
  for (unsigned int i = 0; i < arr->faces.size(); ++i) {
    Face *f = arr->faces[i];
    if (f->cn == n) {
      Region *r = new Region;
      f->traverse(r->boundary);
      regions->push_back(r);
    }
  }
  delete arr;
  delete_segments(segments);
  return regions;
}

void region_segments (Region *region, double *bbox, Segments &segs)
{
  for (unsigned int i = 0; i < region->boundary.size(); ++i) {
    Points &points = region->boundary[i];
    Point t = *(points.end()-1);
    for (unsigned int j = 0; j < points.size(); ++j) {
      Point &h = points[j];
      if (!bbox_segment(t, h, bbox))
	segs.push_back(new Segment(h, t, true));
      t = h;
    }
  }
}

bool bbox_segment (const Point &t, const Point &h, double *bbox)
{
  return t.x == h.x && (t.x == bbox[0] || t.x == bbox[1]) ||
    t.y == h.y && (t.y == bbox[2] || t.y == bbox[3]);
}

Regions * complement (Regions *regions, double *bbox)
{
  Segments segments;
  for (unsigned int i = 0; regions && i < regions->size(); ++i)
    region_segments(regions->at(i), bbox, segments);
  bool brot = bbox[2] == - M_PI && bbox[3] == M_PI;
  form_curves_bbox(brot, bbox, segments);
  Arrangement *arr = arrange(segments);
  Face *f0 = 0;
  Point p0;
  for (unsigned int i = 1; i < arr->faces.size(); ++i) {
    Face *fi = arr->faces[i];
    Region ri;
    fi->traverse(ri.boundary);
    if (ri.boundary[0].size() >= 3 && ri.area() > MINAREA) {
      f0 = fi;
      p0 = ri.interior_point();
      break;
    }
  }
  if (f0 == 0)
    return 0;
  int cn0 = 0;
  for (unsigned int i = 0; regions && i < regions->size(); ++i)
    if (regions->at(i)->contains(p0)) {
      cn0 = 1;
      break;
    }
  arr->set_crossing_numbers(f0, cn0);
  Regions *res = new Regions;
  for (unsigned int i = 0; i < arr->faces.size(); ++i) {
    Face *f = arr->faces[i];
    if (f->cn == 0) {
      Region *r = new Region;
      f->traverse(r->boundary);
      res->push_back(r);
    }
  }
  delete arr;
  delete_segments(segments);
  return res;
}

Pair::~Pair () 
{ 
  delete ra; 
  delete tb;
  delete_regions(regions);
  delete_regions(czone);
  delete_glRegions(glregions);
  delete_glRegions(glczone);
}

void Pair::set_region ()
{
  Point p(a->x, b->x), q;
  double d = regions->at(0)->distance(p, edge, q);
  region = regions->at(0);
  for (unsigned int i = 1; i < regions->size() && d > 0.0; ++i) {
    int edgei[2];
    double di = regions->at(i)->distance(p, edgei, q);
    if (di < d) {
      region = regions->at(i);
      d = di;
      edge[0] = edgei[0];
      edge[1] = edgei[1];
    }
  }
}

bool Pair::set_velocity (bool aflag, bool &blocked)
{
  blocked = false;
  Points &loop = region->boundary[edge[0]];
  int n = loop.size();
  Point p = loop[edge[1]];
  if (p.x == a->x && p.y == b->x) {
    Angle v(a->v, b->v), tb(loop[(edge[1]-1)%n] - p), tf(loop[(edge[1]+1)%n] - p);
    if (tf*tb > 0.0 && v%tf < 0.0)
      if (v%tb < 0.0) {
	a->v = b->v = 0.0;
	blocked = true;
	return true;
      }
      else {
	edge[1] = (edge[1]-1)%n;
	p = loop[edge[1]];
      }
  }
  Point &f = loop[(edge[1]+1)%n];
  Angle nf(p.y - f.y, f.x - p.x);
  double kf = a->v*nf.c + b->v*nf.s;
  if (kf > 0.0) {
    free = true;
    return false;
  }
  if (a->traversed && b->traversed)
    return false;
  if (aflag && fabs(nf.s) < EPSILON) {
    a->v = 0.0;
    return false;
  }
  if (aflag) {
    b->v -= kf/nf.s;
    return true;
  }
  if (fabs(nf.c) < EPSILON) {
    b->v = 0.0;
    return false;
  }
  a->v -= kf/nf.c;
  return true;
}

bool Pair::next_event (SimulateEvent &event)
{
  if (a->v == 0.0 && b->v == 0.0)
    return false;
  Point p(a->x, b->x), v(a->v, b->v);
  event.pair = this;
  if (free) {
    event.edge[0] = edge[0];
    event.edge[1] = edge[1];
    if (!region->ray_intersection(p, v, event.time, event.p, event.edge)) {
      event.time = 0.0;
      region->distance(p, event.edge, event.p);
    }
    return true;
  }
  int n = region->boundary[edge[0]].size();
  Point t, h;
  contact_segment(t, h);
  double k = v%h - v%t;
  if (k == 0.0)
    return false;
  event.edge[0] = edge[0];
  if (k > 0.0) {
    event.p = h;
    event.edge[1] = (edge[1] + 1)%n;
  }
  else {
    event.p = t;
    event.edge[1] = (edge[1] - 1)%n;
  }
  event.time = fabs(v.x) > fabs(v.y) ? (event.p.x - p.x)/v.x 
    : (event.p.y - p.y)/v.y;
  return event.time >= 0.0;
}

void Pair::wrap ()
{
  set_region();
  Points &loop = region->boundary[edge[0]];
  int n = loop.size(), i = edge[1];
  if (fabs(a->x) == M_PI)
    while (loop[i].x == a->x && loop[(i+1)%n].x == a->x) {
      if (a->x == M_PI ? b->x >= loop[(i+1)%n].y : b->x <= loop[(i+1)%n].y)
	i = (i+1)%n;
      else if (a->x == M_PI ? b->x <= loop[i].y : b->x >= loop[i].y)
	i = (i-1)%n;
      else
	break;
    }
  if (fabs(b->x) == M_PI)
    while (loop[i].y == b->x && loop[(i+1)%n].y == b->x) {
      if (b->x == M_PI ? a->x <= loop[(i+1)%n].x : a->x >= loop[(i+1)%n].x)
	i = (i+1)%n;
      else if (b->x == M_PI ? a->x >= loop[i].x : a->x <= loop[i].x)
	i = (i-1)%n;
      else
	break;
    }
  edge[1] = i;
}

void Pair::contact_segment (Point &t, Point &h) const
{
  Points &loop = region->boundary[edge[0]];
  int n = loop.size();
  t = loop[edge[1]];
  h = loop[(edge[1]+1)%n];
}

void Pair::form_czone (double d)
{
  if (czone)
    delete_regions(czone);
  ctriangles.clear();
  delete_glRegions(glczone);
  Part *aa = ra ? ra : a, *bb = tb ? tb : b;
  Part *ap = aa->offset(d), *am = aa->offset(-d),
    *bp = bb->offset(d), *bm = bb->offset(-d);
  double bboxp[4], bboxm[4];
  for (int i = 0; i < 4; ++i)
    bboxp[i] = bboxm[i] = bbox[i];
  Regions *rp = ap->cspace(bp, acc, bboxp, bbox);
  Regions *rm = am->cspace(bm, acc, bboxm, bbox);
  Regions *rpc = complement(rp, bbox);
  RegionsList rl;
  rl.push_back(rpc);
  rl.push_back(rm);
  czone = intersection(rl, bbox);
  triangulate_regions(czone, ctriangles);
  for (unsigned int i = 0; i < czone->size(); ++i)
    glczone.push_back(new glRegion(czone->at(i)->boundary, 0.0, 0.0));
  delete_regions(rp);
  delete_regions(rm);
  delete_regions(rpc);
  delete ap;
  delete am;
  delete bp;
  delete bm;
}

bool Pair::path (const Point &p, const Point &q, Points &res,
		 bool optimize) const
{
  Region *rp = enclosing_region(p);
  if (!rp)
    return false;
  return rp->path(p, q, res);
}

Region * Pair::enclosing_region (const Point &p) const
{
  if (!regions)
    return 0;
  for (unsigned int i = 0; i < regions->size(); ++i)
    if (regions->at(i)->contains(p))
      return regions->at(i);
  return 0;
}

Pair * form_pair (Part *a, Part *b, double acc, double *bbox, bool usebbox)
{
  if (!a->rot && (b->rot || a->t.c != 1.0)) {
    Part *temp = a;
    a = b;
    b = temp;
  }
  if (!(a->rot || b->rot || a->t.c == 1.0 && b->t.s == 1.0))
    return 0;
  Point o(0.0, 0.0);
  Part *ra = (!a->rot || b->rot || b->t.c == 1.0) ? 0 : a->move(- b->t, o);
  Part *tb = b->rot ? 0 : 
    (a->rot ? b->move(- b->t, b->o - a->o) : b->move(0.0, b->o - a->o));
  Pair *pair = new Pair(a, b, ra, tb, acc);
  Part *pa = ra ? ra : a, *pb = tb ? tb : b;
  pair->regions = pa->cspace(pb, acc, pair->bbox, bbox, usebbox);
  for (unsigned int i = 0; i < pair->regions->size(); ++i)
    pair->glregions.push_back(new glRegion(pair->regions->at(i)->boundary, 
					   0.0, 0.0));
  return pair;
}

void triangulate_regions (Regions *regions, Triangles &triangles)
{
  if (regions)
    for (unsigned int i = 0; i < regions->size(); ++i)
      triangulate(regions->at(i)->boundary, triangles);
}

System::~System ()
{
  for (unsigned int i = 0; i < parts.size(); ++i)
    delete parts[i];
  for (unsigned int i = 0; i < pairs.size(); ++i)
    delete pairs[i];
}

void System::read (istream &istr)
{
  int type, version, nparts, npairs, nlinkages, nfpairs;
  istr >> type >> version >> nparts >> npairs >> nlinkages >> nfpairs;
  for (int i = 0; i < nparts; ++i) {
    Part *part = new Part;
    part->read(istr);
    parts.push_back(part);
  }
  for (int i = 0; i < npairs; ++i) {
    int ia, ib;
    double acc;
    istr >> ia >> ib >> acc;
    Pair *pair = form_pair(parts[ia], parts[ib], acc);
    if (pair)
      pairs.push_back(pair);
  }
}

void System::read_motions (istream &istr, Motions &motions, double *c0,
			   double *pb) const
{
  int nmotions, nm, aj;
  double vj, dummy;
  istr >> nmotions;
  for (int i = 0; i < nmotions; ++i) {
    Motion motion;
    istr >> motion.time >> nm;
    for (int j = 0; j < nm; ++j) {
      istr >> aj >> vj;
      motion.elts.push_back(MotionElt(parts[aj], vj));
    }
    motions.push_back(motion);
  }
  for (unsigned int i = 0; i < parts.size(); ++i)
    istr >> c0[i];
  istr >> dummy >> dummy;
  for (int i = 0; i < 4; ++i)
    istr >> pb[i];
  istr >> dummy;
}

void System::simulate (double *c0, const Motions &motions, Trajectory &traj)
{
  double t = 0.0;
  set_state(c0);
  write_state(t, traj);
  for (unsigned int i = 0; i < pairs.size(); ++i)
    if (!pairs[i]->regions || pairs[i]->regions->empty())
      return;
  set_regions();
  for (unsigned int i = 0; i < motions.size(); ++i) {
    const Motion &m = motions[i];
    double ti = 0.0;
    int n = 0;
    while (ti < m.time && n < 100) {
      if (!set_velocity(m))
	break;
      SimulateEvent event;
      if (next_event(event) && ti + event.time < m.time) {
	n = event.time < EPSILON ? n + 1 : 0;
	ti += event.time;
	update_state(event);
	write_state(t + ti, traj);
	if (wrap())
	  write_state(t + ti, traj);
      }
      else {
	update_state(m.time - ti);
	ti = m.time;
	write_state(t + ti, traj);
      }
    }
    t += ti;
  }
}

void System::set_state (double *state)
{
  for (unsigned int i = 0; i < parts.size(); ++i)
    parts[i]->x = state[i];
}

void System::write_state (double t, Trajectory &traj) const
{
  traj.push_back(t);
  for (unsigned int i = 0; i < parts.size(); ++i)
    traj.push_back(parts[i]->x);
}

void System::set_regions ()
{
  for (unsigned int i = 0; i < pairs.size(); ++i) {
    pairs[i]->set_region();
    pairs[i]->free = true;
  }
}

bool System::set_velocity (const Motion &motion)
{
  for (unsigned int i = 0; i < parts.size(); ++i) {
    parts[i]->v = 0.0;
    parts[i]->traversed = false;
  }
  for (unsigned int i = 0; i < pairs.size(); ++i)
    pairs[i]->traversed = false;
  for (unsigned int i = 0; i < motion.elts.size(); ++i)
    motion.elts[i].part->v = motion.elts[i].v;
  for (unsigned int i = 0; i < motion.elts.size(); ++i)
    if (!set_velocity(motion.elts[i].part))
      return false;
  return true;
}

bool System::set_velocity (Part *p)
{
  p->traversed = true;
  bool blocked;
  for (unsigned int i = 0; i < pairs.size(); ++i)
    if (pairs[i]->traversed || pairs[i]->free)
      ;
    else if (pairs[i]->a == p) {
      pairs[i]->traversed = true;
      if (pairs[i]->set_velocity(true, blocked)) {
	if (blocked) 
	  return false;
	set_velocity(pairs[i]->b);
      }
    }
    else if (pairs[i]->b == p) {
      pairs[i]->traversed = true;
      if (pairs[i]->set_velocity(false, blocked)) {
	if (blocked)
	  return false;
	set_velocity(pairs[i]->a);
      }
    }
  return true;
}

bool System::next_event (SimulateEvent &event) const
{
  bool flag = false;
  for (unsigned int i = 0; i < pairs.size(); ++i) {
    SimulateEvent ievent;
    bool iflag = pairs[i]->next_event(ievent);
    if (iflag && (!flag || ievent.time < event.time)) {
      flag = true;
      event = ievent;
    }
  }
  return flag;
}

void System::update_state (const SimulateEvent &event)
{
  update_state(event.time);
  Pair &pe = *event.pair;
  pe.edge[0] = event.edge[0];
  pe.edge[1] = event.edge[1];
  pe.a->x = event.p.x;
  pe.b->x = event.p.y;
  pe.free = false;
}

void System::update_state (double time)
{
  for (unsigned int i = 0; i < parts.size(); ++i)
    parts[i]->x += parts[i]->v*time;
}

bool System::wrap ()
{
  Pairs wpairs;
  for (unsigned int i = 0; i < parts.size(); ++i) {
    Part *pi = parts[i];
    if (pi->wrap())
      for (unsigned int j = 0; j < pairs.size(); ++j)
	if (pi == pairs[j]->a || pi == pairs[j]->b)
	  wpairs.push_back(pairs[j]);
  }
  for (unsigned int i = 0; i < wpairs.size(); ++i)
    wpairs[i]->wrap();
  return !wpairs.empty();
}

void System::form_profiles (double *z)
{
  if (!parts[0]->profiles.empty())
    return;
  z[0] = BIGNUMBER;
  z[1] = - BIGNUMBER;
  for (unsigned int i = 0; i < parts.size(); ++i) {
    parts[i]->form_profiles(0.01, z);
    parts[i]->triangulate();
  }
  z[0] -= 1.0;
  z[1] += 1.0;
}

void System::form_pair_trajectories (Trajectory &traj)
{
  int np = parts.size(), n = traj.size()/(np + 1);
  for (unsigned int i = 0; i < pairs.size(); ++i) {
    Pair *p = pairs[i];
    bool ra = p->a->rot, rb = p->b->rot;
    int ia = find(parts.begin(), parts.end(), p->a) - parts.begin();
    int ib = find(parts.begin(), parts.end(), p->b) - parts.begin();
    p->gltraj.nvb.push_back(0);
    p->gltraj.vb = new double [3*n];
    double *vbp = p->gltraj.vb, ox = 0.0, oy = 0.0;
    for (int j = 0; j < n; ++j, vbp += 3) {
      double x = traj[j*(np+1)+ia+1], y = traj[j*(np+1)+ib+1];
      if (ra && (fabs(ox) == M_PI || fabs(x) == M_PI) && fabs(ox - x) > 6.28 ||
	  rb && (fabs(oy) == M_PI || fabs(y) == M_PI) && fabs(oy - y) > 6.28)
	p->gltraj.nvb.push_back(j);
      vbp[0] = x;
      vbp[1] = y;
      vbp[2] = 0.0;
      ox = x;
      oy = y;
    }
    p->gltraj.nvb.push_back(n);
  }
}

void System::form_czone (double d)
{
  for (unsigned int i = 0; i < pairs.size(); ++i)
    pairs[i]->form_czone(d);
}
